from __future__ import annotations

import functools
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    Literal,
    TypeVar,
    Union,
    cast,
    overload,
)

import higlass_schema as hgs
from pydantic import BaseModel

import higlass._display as display
import higlass._utils as utils

__all__ = [
    "EnumTrack",
    "HeatmapTrack",
    "IndependentViewportProjectionTrack",
    "CombinedTrack",
    "PluginTrack",
    "TrackT",
    "View",
    "ViewT",
    "Viewconf",
    "concat",
    "hconcat",
    "vconcat",
    "track",
    "view",
    "combine",
    "divide",
    "lock",
]

if TYPE_CHECKING:
    from higlass.server import TilesetResource

## Mixins

# Can't figure out a good way to type these classes.
# We ignoring the errors in the parameter signature
# mean we can get good type information within the
# function and return types are inferred for end users.


class _PropertiesMixin:
    def properties(
        self: utils.ModelT,  # type: ignore
        inplace: bool = False,
        **fields,  # type: ignore
    ) -> utils.ModelT:  # type: ignore
        """Configures top-level properties.

        Updates top-level properties for a Track, View, or Viewconf. This
        is really a convenience to allow method chaining/derived objects
        to be created without multiple lines of mutating the classes. For
        example,

        >>> view = hg.view(hg.track("heatmap"))
        >>> updated_view = view.properties(zoomFixed=True)
        >>> assert view.uid != updated_view.uid
        >>> assert view.zoomFixed is None
        >>> assert updated_view.zoomFixed is True

        inplace : bool, optional
            Whether to modify the existing track in place or return
            a new track with the options applied (default: `False`).

        **fields : dict
            The updated properties for .

        Returns
        -------
        track : A track with the the newly specified track options.

        """
        model = self if inplace else utils.copy_unique(self)
        for k, v in fields.items():
            setattr(model, k, v)
        return model


class _OptionsMixin:
    def opts(
        self: TrackT,  # type: ignore
        inplace: bool = False,
        **options,
    ) -> TrackT:  # type: ignore
        """Configures options for a Track.

        A convenience method to update `track.options`.

        Parameters
        ----------
        inplace : bool, optional
            Whether to modify the existing track in place or return
            a new track with the options applied (default: `False`).

        **options : dict
            Options to pass down to the underlying track object.

        Returns
        -------
        track : A track with the the newly specified track options.

        Examples
        --------
        >>> track = hg.track("heatmap")
        >>> derived = track.opts(colorRange=["rgba(255,255,255,1)", "rgba(0,0,0,1)"])
        >>> assert track.options is None
        >>> assert isinstance(derived.options["colorRange"], list)

        """
        track = self if inplace else utils.copy_unique(self)
        if track.options is None:
            track.options = {}
        track.options.update(options)
        return track


class _TilesetMixin:
    def tileset(
        self: TrackT,  # type: ignore
        tileset: TilesetResource,
        inplace: bool = False,
    ) -> TrackT:  # type: ignore
        """Replace or add a tileset to a Track.

        A convenience method to update a track with a tileset.

        Parameters
        ----------
        tileset : TilesetResource
            A tileset resource returned from `hg.server`.

        inplace : bool, optional
            Whether to modify the existing track in place or return
            a new track (default: `False`)

        Returns
        -------
        track : A track with the bound tileset.

        """
        track = self if inplace else utils.copy_unique(self)
        track.server = tileset.server
        track.tilesetUid = tileset.tileset.uid
        return track


## Extend higlass-schema classes


class EnumTrack(hgs.EnumTrack, _OptionsMixin, _PropertiesMixin, _TilesetMixin):
    """Represents a generic track."""

    ...


class HeatmapTrack(hgs.HeatmapTrack, _OptionsMixin, _PropertiesMixin, _TilesetMixin):
    """Represets a specialized heatmap track."""

    ...


class IndependentViewportProjectionTrack(
    hgs.IndependentViewportProjectionTrack,
    _OptionsMixin,
    _PropertiesMixin,
    _TilesetMixin,
):
    """Represents a view-independent viewport projection track."""

    ...


class CombinedTrack(hgs.CombinedTrack, _OptionsMixin, _PropertiesMixin, _TilesetMixin):
    """Represents a track combining multiple tracks."""

    ...


class PluginTrack(hgs.BaseTrack, _OptionsMixin, _PropertiesMixin, _TilesetMixin):
    """Represents an unknown plugin track."""

    plugin_url: ClassVar[str]


Track = Union[
    HeatmapTrack,
    IndependentViewportProjectionTrack,
    EnumTrack,
    CombinedTrack,
    PluginTrack,
]

TrackT = TypeVar("TrackT", bound=Track)


class View(hgs.View[TrackT], _PropertiesMixin, Generic[TrackT]):
    """Represets a HiGlass View that is generic over the inner tracks."""

    def domain(
        self,
        x: hgs.Domain | None = None,
        y: hgs.Domain | None = None,
        inplace: bool = False,
    ):
        """Configures the view x and/or y domain(s).

        Parameters
        ----------
        x : tuple[float, float], optional
            The x domain for the view (default: `None`)

        y : tuple[float, float], optional
            The y domain for the view (default: `None`)

        inplace : bool, optional
            Whether to modify this view inplace. If `False` (default),
            a new view is created and returned with the domains
            applied (default: `False`)

        Returns
        -------
        view : a View with the updated x & y domains

        """
        view = self if inplace else utils.copy_unique(self)
        if x is not None:
            view.initialXDomain = x
        if y is not None:
            view.initialYDomain = y
        return view

    def __or__(self, other: View[TrackT] | Viewconf[TrackT]):
        """Horizontally concatenate with another view or viewconf.

        A convenience method for `hg.hconcat`.

        Parameters
        ----------
        other : View | Viewconf
            The other view or viewconf to combine with.

        Returns
        -------
        viewconf : A combined Viewconf

        """
        return hconcat(self, other)

    def __truediv__(self, other: View[TrackT] | Viewconf[TrackT]):
        """Vertically concatenate two Views.

        A convenience method for `hg.vconcat`.

        Parameters
        ----------
        other : View | Viewconf
            The other view or viewconf to combine with.

        Returns
        -------
        viewconf : A combined Viewconf

        """
        return vconcat(self, other)

    def clone(self):
        """Clone the the View instance.

        Inserts a unique uuid so the view can be uniquely identified in the front end.

        Returns
        -------
        view : A copy of the original view with a new uuid

        """
        return utils.copy_unique(self)

    def viewconf(self, **kwargs):
        """Consumes the current View into a top-level view config.

        Returns
        -------
        viewconf : A top-level HiGlass Viewconf.

        """
        return Viewconf[TrackT](views=[self], **kwargs)

    def _repr_mimebundle_(self, include=None, exclude=None):
        """ "Displays the view in an IPython environment."""
        return self.viewconf()._repr_mimebundle_(include, exclude)

    def widget(self, **kwargs):
        """Create a Jupyter Widget display for this view.

        Casts the view into a top-level view config which can be displayed.

        Returns
        -------
        widget : A HiGlassWidget instance
        """
        return self.viewconf().widget(**kwargs)

    def project(
        self,
        view: View,
        on: Literal["center", "top", "bottom", "left", "right"] = "center",
        inplace: bool = False,
        **kwargs,
    ):
        """Project a different viewport onto this view.

        Creates a viewport-projection track whos bounds are synchonrized
        with an existing, separate view.

        Parameters
        ----------
        view : View
            The view with the desired viewport to project.

        on : Literal["center", "top", "bottom", "left", "right"], optional
            Where to project the viewport on this view (default: "center").

        inplace : bool
            Whether to modify this view inplace, or alternatively return
            a new view (default: `False`).

        **kwargs : dict
            Additional parameters for the created viewport-projection track.

        Returns
        -------
        view : A view with a new viewport-projection track.
        """
        new_view = self if inplace else utils.copy_unique(self)

        # projection track type from position
        if on == "center":
            track_type = "viewport-projection-center"
        elif on == "top" or on == "bottom":
            track_type = "viewport-projection-horizontal"
        elif on == "left" or on == "right":
            track_type = "viewport-projection-vertical"
        else:
            raise ValueError("Not possible")

        if getattr(new_view.tracks, on) is None:
            setattr(new_view.tracks, on, [])

        trk = track(type_=track_type, fromViewUid=view.uid, **kwargs)
        getattr(new_view.tracks, on).append(trk)

        return new_view


ViewT = TypeVar("ViewT", bound=View)


def gather_plugin_urls(views: list[ViewT]) -> list[str]:
    """Inspect tracks and extract plugin urls to inject into HTML."""
    plugin_urls = {}
    for view in views:
        for _, track in view.tracks:
            if isinstance(track, PluginTrack):
                plugin_urls[track.type] = track.plugin_url
    return list(plugin_urls.values())


class Viewconf(hgs.Viewconf[View[TrackT]], _PropertiesMixin, Generic[TrackT]):
    """Represents a top-level viewconfig, or a complete HiGlass visualization."""

    def _repr_mimebundle_(self, include=None, exclude=None):
        """ "Displays the view config in an IPython environment."""
        renderer = display.renderers.get()
        plugin_urls = [] if self.views is None else gather_plugin_urls(self.views)
        return renderer(self.dict(), plugin_urls=plugin_urls)

    def widget(self):
        """Create a Jupyter Widget display for this view config."""
        from higlass_widget import HiGlassWidget

        return HiGlassWidget(self.dict())

    @classmethod
    def from_url(cls, url: str, **kwargs):
        """Load a viewconf via URL and construct a Viewconf.

        Makes an HTTP request.

        Parameters
        ----------
        url : str
            The URL for a JSON HiGlass view config.

        **kwargs : dict
            Options for the `urllib.Request` instance.

        Returns
        -------
        viewconf : The parsed view config

        """
        import urllib.request as urllib

        request = urllib.Request(url, **kwargs)
        with urllib.urlopen(request) as response:
            raw = response.read()

        return cls.parse_raw(raw)

    def locks(
        self,
        *locks: hgs.Lock | hgs.ValueScaleLock,
        zoom: list[hgs.Lock] | hgs.Lock | None = None,
        location: list[hgs.Lock] | hgs.Lock | None = None,
        value_scale: list[hgs.ValueScaleLock] | hgs.ValueScaleLock | None = None,
        inplace: bool = False,
    ):
        """Specify view locks.

        Parameters
        ----------

        *locks : hgs.Lock | hgs.ValueScaleLock, optional
            A set of either zoom/location or value scale locks to add. Instances
            of `hgs.Lock` will synchronize _both_ zoom and location for the views
            included in the lock. If you'd like to synchronize just zoom or location,
            use the explicit keyword arguments below.

        zoom : hgs.Lock | list[hgs.Lock], optional
            A lock or set of locks to synchronize only the zoom.

        location : hgs.Lock | list[hgs.Lock], optional
            A lock or set of locks to synchronize only the location.

        value_scale : hgs.ValueScaleLock | list[hgs.ValueScaleLock], optional
            A single or set of value-scale locks.

        inplace : bool, optional
            Whether to modify the viewconf in place or return a new viewconf
            with the locks applied (default: `False`)

        Returns
        -------

        viewconf : A Viewconf with the synchronized views.


        Examples
        --------

        >>> view1 = hg.view(hg.track("heatmap"))
        >>> view2 = hg.view(hg.track("heatmap"))
        >>> # create an abstract lock for two views
        >>> view_lock = hg.lock(view1, view2)
        >>> # lock location & zoom
        >>> (view1 | view2).lock(view_lock)
        >>> # lock only zoom
        >>> (view1 | view2).lock(zoom=view_lock)
        >>> # lock only location
        >>> (view1 | view2).lock(location=view_lock)

        """
        conf = self if inplace else utils.copy_unique(self)

        zoom = utils.ensure_list(zoom)
        location = utils.ensure_list(location)
        value_scale = utils.ensure_list(value_scale)

        shared_locks: list[hgs.Lock] = []
        for lock in locks:
            if isinstance(lock, hgs.Lock):
                shared_locks.append(lock)
            else:
                value_scale.append(lock)

        zoom.extend(shared_locks)
        location.extend(shared_locks)

        if conf.zoomLocks is None:
            conf.zoomLocks = hgs.ZoomLocks()

        for lock in zoom:
            assert isinstance(lock.uid, str)
            conf.zoomLocks.locksDict[lock.uid] = lock
            for vuid, _ in lock:
                conf.zoomLocks.locksByViewUid[vuid] = lock.uid

        if conf.locationLocks is None:
            conf.locationLocks = hgs.LocationLocks()

        for lock in location:
            assert isinstance(lock.uid, str)
            conf.locationLocks.locksDict[lock.uid] = lock
            for vuid, _ in lock:
                conf.locationLocks.locksByViewUid[vuid] = lock.uid

        if conf.valueScaleLocks is None:
            conf.valueScaleLocks = hgs.ValueScaleLocks()

        for lock in value_scale:
            assert isinstance(lock.uid, str)
            conf.valueScaleLocks.locksDict[lock.uid] = lock
            for vuid, _ in lock:
                conf.valueScaleLocks.locksByViewUid[vuid] = lock.uid

        return conf

    def __or__(self, other: View[TrackT] | Viewconf[TrackT]) -> Viewconf[TrackT]:
        """Horizontally concatenate with another view or viewconf.

        A convenience method for `hg.hconcat`.

        Parameters
        ----------
        other : View | Viewconf
            The other view or viewconf to combine with.

        Returns
        -------
        viewconf : A combined Viewconf

        """
        return hconcat(self, other)

    def __truediv__(self, other: View[TrackT] | Viewconf[TrackT]) -> Viewconf[TrackT]:
        """Vertically concatenate with another view or viewconf.

        A convenience method for `hg.hconcat`.

        Parameters
        ----------
        other : View | Viewconf
            The other view or viewconf to combine with.

        Returns
        -------
        viewconf : A combined Viewconf

        """
        return vconcat(self, other)


TrackTA = TypeVar("TrackTA", bound=Track)
TrackTB = TypeVar("TrackTB", bound=Track)


def concat(
    method: Literal["horizontal", "vertical"],
    a: View[TrackTA] | Viewconf[TrackTA],
    b: View[TrackTB] | Viewconf[TrackTB],
) -> Viewconf[TrackTA | TrackTB]:
    """Concatenate two views or separate viewconfs together.

    Uses the layout of one view or the total bounds of all views in one
    viewconf to offset the other view/viewconf.

    Parameters
    ----------
    method : Literal["horizontal", "vertical"]
        How to concatenate views/viewconfs.

    a : View | Viewconf
        A view or viewconf to combine.

    b : View | Viewconf
        The other view or viewconf to combine.

    Returns
    -------
    viewconf : A combined viewconf containing multiple views.

    """
    a = a.viewconf() if isinstance(a, View) else a
    assert a.views is not None

    b = b.viewconf() if isinstance(b, View) else b
    assert b.views is not None

    if method == "vertical":

        def mapper(view):
            return view.layout.y + view.layout.h

        field = "y"
    elif method == "horizontal":

        def mapper(view):
            return view.layout.x + view.layout.w

        field = "x"
    else:
        raise ValueError("concat method must be 'vertical' or 'horizontal'.")

    # gather views and adjust layout
    views = [v.copy(deep=True) for v in b.views]
    offset = 0 if a.views is None else max(map(mapper, a.views))
    for view in views:
        curr = getattr(view.layout, field)
        setattr(view.layout, field, curr + offset)

    a.views.extend(views)  # type: ignore

    # merge locks
    for lockattr in ["zoomLocks", "valueScaleLocks", "locationLocks"]:
        locks = getattr(b, lockattr)
        if locks:
            if getattr(a, lockattr) is None:
                setattr(a, lockattr, locks.copy(deep=True))
            else:
                getattr(a, lockattr).locksByViewUid.update(locks.locksByViewUid)
                getattr(a, lockattr).locksDict.update(locks.locksDict)

    return cast(Viewconf[Union[TrackTA, TrackTB]], a)


hconcat = functools.partial(concat, "horizontal")

vconcat = functools.partial(concat, "vertical")


## Top-level functions to easily create tracks,

# TODO: register plugins globally to work here?


class _TrackCreator(BaseModel):
    """Create track instances from their track type.

    Used internally by `hg.track` to leverage pydantic's ability to get
    the appropriate base model by the track type.

    Example:
    -------
    >>> assert isinstance(_TrackCreator(type="heatmap").__root__, HeatmapTrack)
    """

    __root__: Track


@overload
def track(type_: hgs.EnumTrackType, uid: str | None = None, **kwargs) -> EnumTrack:
    ...


@overload
def track(type_: Literal["heatmap"], uid: str | None = None, **kwargs) -> HeatmapTrack:
    ...


@overload
def track(type_: str, uid: str | None = None, **kwargs) -> PluginTrack:
    ...


def track(
    type_: hgs.EnumTrackType | Literal["heatmap"] | str,
    uid: str | None = None,
    **kwargs,
) -> Track:
    """Create a HiGlass track.

    Parameters
    ----------
    type_ : str
        The track type to create

    uid: str, optional
        A unique id for the track. If unspecified (default), a unique id is given
        internally.

    **kwargs: dict
        Other top-level track properties.

    Returns
    -------
    track :  An instance of an hg.Track
    """
    if uid is None:
        uid = utils.uid()
    data = dict(type=type_, uid=uid, **kwargs)
    return _TrackCreator.parse_obj(data).__root__


def view(
    *_tracks: TrackT | tuple[TrackT, utils.TrackPosition] | hgs.Tracks[TrackT],
    x: int = 0,
    y: int = 0,
    width: int = 12,
    height: int = 6,
    tracks: hgs.Tracks[TrackT] | None = None,
    layout: hgs.Layout | None = None,
    uid: str | None = None,
    **kwargs,
) -> View[TrackT]:
    """Create a HiGlass view from multiple tracks.

    Parameters
    ----------
    *_tracks : Track | tuple[Track, str] | hgs.Tracks[Track]
        The tracks to include in the view. Can be 1) separate track objects
        (position inferred), 2) explicit (Track, position) tuples, or 3.) a
        `higlass_schema.Tracks` Object specifying all tracks and positions.

    x : int, optional
        The x position of the view in the HiGlass grid (default: 0). Used to
        created a Layout if none is provided.

    y : int, optional
        The y position of the view in the HiGlass grid (default: 0). Used to
        created a Layout if none is provided.

    width : int, optional
        The width of the view (default: 12). Used to created a Layout if none is
        provided.

    height : int, optional
        The height of the view (default: 6). Used to created a Layout if none is
        provided.

    layout : hgs.Layout, optional
        An explicit layout for the view (default: None). If provided the `x`, `y`,
        `width`, and `height` parameters are ignored.

    uid: str, optional
        A unique id for the view. If unspecified (default), a unique id is automatically
        generated internally.

    **kwargs: dict
        Other top-level view properties.

    Returns
    -------
    view :  An instance of an hg.View
    """
    if layout is None:
        layout = hgs.Layout(x=x, y=y, w=width, h=height)
    else:
        layout = hgs.Layout(**layout.dict())

    if tracks is None:
        data = defaultdict(list)
    else:
        data = defaultdict(list, tracks.dict())

    for track in _tracks:
        if isinstance(track, hgs.Tracks):
            track = track.dict()
            for position, track_list in track.items():
                data[position].extend(track_list)
        else:
            if isinstance(track, tuple):
                track, position = track
            else:
                position = utils.track_default_position.get(track.type)
                if position is None:
                    raise ValueError("No default track type")

            data[position].append(track)

    if uid is None:
        uid = utils.uid()

    return View[TrackT](
        layout=layout,
        tracks=hgs.Tracks[TrackT](**data),
        uid=uid,
        **kwargs,
    )


def combine(t1: Track, t2: Track, uid: str | None = None, **kwargs) -> CombinedTrack:
    """A utility to create a CombinedTrack from two other tracks.

    Parameters
    ----------

    t1 : hg.Track
        The first of two tracks to combine.

    t2 : hg.Track
        The second of two tracks to combine.

    uid : str, optional
        A unique id for the newly created CombinedTrack.

    **kwargs : dict
        Additional properties to pass to the CombinedTrack constructor.

    Returns
    -------

    combined_track : An CombinedTrack with two children tracks.

    """
    if uid is None:
        uid = utils.uid()

    if isinstance(t1, CombinedTrack):
        copy = CombinedTrack(**t1.dict())
        copy.contents.append(t2.__class__(**t2.dict()))
        for key, val in kwargs.items():
            setattr(copy, key, val)
        return copy

    return CombinedTrack(
        type="combined",
        uid=uid,
        contents=[track.__class__(**track.dict()) for track in (t1, t2)],
        **kwargs,
    )


T = TypeVar("T", bound=Union[EnumTrack, HeatmapTrack])


def divide(t1: T, t2: T, **kwargs) -> T:
    """A utility to create a divided track.

    Tracks must the the same type.

    Parameters
    ----------

    t1 : hg.EnumTrack | hg.HeatmapTrack
        The first track to divide by the other.

    t2 : hg.EnumTrack | hg.HeatmapTrack
        The other track.

    **kwargs : dict
        Additional properties to pass to the newly created track.

    Returns
    -------

    divided_track : An identical track with "divided" data sources.

    """
    assert t1.type == t2.type, "divided tracks must be same type"
    assert isinstance(t1.tilesetUid, str)
    assert isinstance(t1.server, str)

    assert isinstance(t2.tilesetUid, str)
    assert isinstance(t2.server, str)

    copy = utils.copy_unique(t1)
    copy.tilesetUid = None
    copy.server = None
    copy.data = hgs.Data(
        type="divided",
        children=[
            {
                "tilesetUid": track.tilesetUid,
                "server": track.server,
            }
            for track in (t1, t2)
        ],
    )
    # overrides
    for key, val in kwargs.items():
        setattr(copy, key, val)
    return copy


@overload
def lock(*views: View, uid: str | None = None) -> hgs.Lock:
    """Create an abstract lock linking two or more views.

    Parameters
    ----------

    *views : View
        Two or more views to synchronize with this lock.

    uid : str, optional
        An (optional) unique id for this lock.

    Returns
    -------

    lock : An abstract lock linking two or more views.

    """
    ...


@overload
def lock(
    *pairs: tuple[View, Track],
    uid: str | None = None,
    ignoreOffScreenValues: bool | None = None,
) -> hgs.ValueScaleLock:
    """Create an abstract value-scale lock linking one or more (View, Track) pairs.

    Parameters
    ----------

    *pairs : tuple[View, Track]
        One or more (View, Track) pairs specifying a Track whos values should
        be used to scale all tracks within the View.

    uid : str, optional
        An (optional) unique id for this lock.

    ignoreOffScreenValues : bool, optional
        Whether to ignore off screen values when scaling the view.

    Returns
    -------

    lock : A value-scale lock defining how the tracks in a View should be scaled by
        by a particular track.

    """
    ...


def lock(*data, uid: str | None = None, **kwargs):
    """Create an abstract lock or value-scale lock.

    Overloaded to either return a `hgs.Lock` or `hgs.ValueScaleLock` depending on
    the arguments. See `Viewconf.locks` for how to use the created locks in
    your top-level view config.

    Examples
    --------

    Create an abstract lock linking two or more separate views.

    >>> view1 = hg.view(hg.track("heatmap"))
    >>> view2 = hg.view(hg.track("heatmap"))
    >>> lock = hg.lock(view1, view2)

    Create a value-scale lock, scaling two views by the values of a particular track.

    >>> t1 = hg.track("heatmap")
    >>> t2 = hg.track("heatmap")
    >>> view1 = hg.view(t1, t2)
    >>> view2 = hg.view(t1, t2)
    >>> value_scale_lock = hg.lock((view1, t2), (view2, t2))

    """
    assert len(data) >= 1
    if uid is None:
        uid = utils.uid()
    if isinstance(data[0], View):
        lck = hgs.Lock(uid=uid)
        for view in data:
            assert isinstance(view.uid, str)
            setattr(lck, view.uid, (1, 1, 1))
        return lck
    else:
        lck = hgs.ValueScaleLock(uid=uid, **kwargs)
        for view, track in data:
            assert isinstance(view.uid, str)
            assert isinstance(track.uid, str)
            vtuid = f"{view.uid}.{track.uid}"
            setattr(lck, vtuid, {"track": track.uid, "view": view.uid})
        return lck



#!/usr/bin/env python3


from docutils import nodes
from docutils.parsers.rst import Directive, directives
from higlass import __version__
from pkg_resources import parse_version

# -*- coding: utf-8 -*-
#
# higlass documentation build configuration file, created by
# sphinx-quickstart on Mon Jul  3 16:40:45 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.imgmath",
    "sphinx.ext.todo",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "HiGlass"
copyright = "2017-2023 HiGlass Authors"
author = "HiGlass Authors"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = parse_version(__version__).base_version
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
# html_extra_path = ['examples']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "higlass_theme"
html_theme_path = ["."]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"sidebar_collapse": False}


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "donate.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "higlassdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "higlass.tex", "HiGlass Documentation", "Peter Kerpedjiev", "manual")
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "higlass", "HiGlass Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "higlass",
        "HiGlass Documentation",
        author,
        "HiGlass",
        "A visual explorer for large genomic data.",
        "Miscellaneous",
    )
]


# -- Options for embedding YouTube and Vimeo videos -----------------------

"""
    ReST directive for embedding Youtube and Vimeo videos.
    There are two directives added: ``youtube`` and ``vimeo``. The only
    argument is the video id of the video to include.
    Both directives have three optional arguments: ``height``, ``width``
    and ``align``. Default height is 281 and default width is 500.
    Example::
        .. youtube:: anwy2MPT5RE
            :height: 315
            :width: 560
            :align: left
    :copyright: (c) 2012 by Danilo Bargen.
    :license: BSD 3-clause
"""


def align(argument):
    """Conversion function for the "align" option."""
    return directives.choice(argument, ("left", "center", "right"))


class IframeVideo(Directive):
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "height": directives.nonnegative_int,
        "width": directives.nonnegative_int,
        "align": align,
        "css": directives.unchanged,
    }
    default_width = 500
    default_height = 281

    def run(self):
        self.options["video_id"] = directives.uri(self.arguments[0])
        self.options["caption"] = "</br>".join(self.content)
        if not self.options.get("width"):
            self.options["width"] = self.default_width
        if not self.options.get("height"):
            self.options["height"] = self.default_height
        if not self.options.get("align"):
            self.options["align"] = "left"
        if not self.options.get("css"):
            self.options["css"] = ""
        return [nodes.raw("", self.html % self.options, format="html")]


class Youtube(IframeVideo):
    html = '<div class="figure align-%(align)s" style="%(css)s">\
    <iframe src="http://www.youtube.com/embed/%(video_id)s" \
    width="%(width)u" height="%(height)u" frameborder="0" \
    webkitAllowFullScreen mozallowfullscreen allowfullscreen></iframe>\
    <p class="caption"><span class="caption-text">%(caption)s</span>\
    </p></div>'


class Vimeo(IframeVideo):
    html = '<div class="figure align-%(align)s" style="%(css)s">\
    <iframe src="http://player.vimeo.com/video/%(video_id)s" \
    width="%(width)u" height="%(height)u" frameborder="0" \
    webkitAllowFullScreen mozallowfullscreen allowFullScreen></iframe>\
    <p class="caption"><span class="caption-text">%(caption)s</span>\
    </p></div>'


def setup(builder):
    directives.register_directive("youtube", Youtube)
    directives.register_directive("vimeo", Vimeo)


autodoc_typehints = "none"



from __future__ import annotations

import itertools
import os
import weakref
from dataclasses import dataclass
from typing import MutableMapping

import starlette.applications
import starlette.middleware.cors
import starlette.requests
import starlette.responses
import starlette.routing

from higlass._utils import TrackType, datatype_default_track
from higlass.api import track
from higlass.tilesets import LocalTileset

from ._background_server import BackgroundServer


@dataclass(frozen=True)
class TilesetResource:
    tileset: LocalTileset
    provider: TilesetProvider

    @property
    def server(self) -> str:
        return f"{self.provider.url}/api/v1/"

    def track(self, type_: TrackType | None = None, **kwargs):
        # use default track based on datatype if available
        if type_ is None:
            if self.tileset.datatype is None:
                raise ValueError("No default track for tileset")
            else:
                type_ = datatype_default_track[self.tileset.datatype]  # type: ignore
        t = track(
            type_=type_,  # type: ignore
            server=self.server,
            tilesetUid=self.tileset.uid,
            **kwargs,
        )
        if self.tileset.name:
            t.opts(name=self.tileset.name, inplace=True)
        return t


def get_list(query: str, field: str) -> list[str]:
    """Parse chained query params into list.
    >>> get_list("d=id1&d=id2&d=id3", "d")
    ['id1', 'id2', 'id3']
    >>> get_list("d=1&e=2&d=3", "d")
    ['1', '3'].
    """
    kv_tuples = [x.split("=") for x in query.split("&")]
    return [v for k, v in kv_tuples if k == field]


# adapted from https://github.com/higlass/higlass-python/blob/b3be6e49cbcab6be72eb0ad65c68a286161b8682/higlass/server.py#L169-L199
def create_tileset_route(tileset_resources: MutableMapping[str, LocalTileset]):
    def tileset_info(request: starlette.requests.Request):
        uids = get_list(request.url.query, "d")
        info = {
            uid: tileset_resources[uid].info()
            if uid in tileset_resources
            else {"error": f"No such tileset with uid: {uid}"}
            for uid in uids
        }
        return starlette.responses.JSONResponse(info)

    def tiles(request: starlette.requests.Request):
        requested_tids = set(get_list(request.url.query, "d"))
        if not requested_tids:
            return starlette.responses.JSONResponse(
                {"error": "No tiles requested"}, 400
            )

        tiles: list = []
        for uid, tids in itertools.groupby(
            iterable=sorted(requested_tids), key=lambda tid: tid.split(".")[0]
        ):
            tileset_resource = tileset_resources.get(uid)
            if not tileset_resource:
                return starlette.responses.JSONResponse(
                    {"error": f"No tileset found for requested uid: {uid}"}, 400
                )
            tiles.extend(tileset_resource.tiles(list(tids)))
        data = {tid: tval for tid, tval in tiles}
        return starlette.responses.JSONResponse(data)

    def chromsizes(request: starlette.requests.Request):
        """Return chromsizes for given tileset id as TSV."""
        uid = request.query_params.get("id")
        if uid is None:
            return starlette.responses.JSONResponse({"error": "No uid provided."}, 400)
        tileset_resource = tileset_resources[uid]
        info = tileset_resource.info()
        assert "chromsizes" in info, "No chromsizes in tileset info"
        return starlette.responses.PlainTextResponse(
            "\n".join(f"{chrom}\t{size}" for chrom, size in info["chromsizes"])
        )

    return starlette.routing.Mount(
        path="/api/v1",
        routes=[
            starlette.routing.Route("/tileset_info/", endpoint=tileset_info),
            starlette.routing.Route("/tiles/", endpoint=tiles),
            starlette.routing.Route("/chrom-sizes/", endpoint=chromsizes),
        ],
    )


class TilesetProvider(BackgroundServer):
    _tilesets: MutableMapping[str, LocalTileset]
    proxy: bool = False

    def __init__(self, allowed_origins: list[str] | None = None):
        if allowed_origins is None:
            allowed_origins = ["*"]
        self._tilesets = weakref.WeakValueDictionary()
        app = starlette.applications.Starlette(
            routes=[
                create_tileset_route(self._tilesets),
            ]
        )

        # configure cors
        if allowed_origins:
            app.add_middleware(
                starlette.middleware.cors.CORSMiddleware,
                allow_origins=allowed_origins,
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        super().__init__(app)

    @property
    def url(self) -> str:
        if self.proxy:
            return f"/proxy/{self.port}"

        # https://github.com/yuvipanda/altair_data_server/blob/4d6ffcb19f864218c8d825ff2c95a1c8180585d0/altair_data_server/_altair_server.py#L73-L93
        if "JUPYTERHUB_SERVICE_PREFIX" in os.environ:
            urlprefix = os.environ["JUPYTERHUB_SERVICE_PREFIX"]
            return f"{urlprefix}/proxy/{self.port}"

        return f"http://localhost:{self.port}"

    def create(self, tileset: LocalTileset) -> TilesetResource:
        resource = TilesetResource(tileset, provider=self)
        self._tilesets[tileset.uid] = tileset
        self.start()
        return resource



import pytest

from higlass.server import HgServer


@pytest.fixture
def server():
    server = HgServer()
    yield server
    server.reset()


def test_server(server: HgServer):

    with pytest.raises(RuntimeError):
        server.port



from __future__ import annotations

import uuid
from typing import Literal, TypeVar, Union

import higlass_schema as hgs
from pydantic import BaseModel

TrackType = Union[hgs.EnumTrackType, Literal["heatmap"]]
TrackPosition = Literal["center", "top", "left", "bottom", "center", "whole", "gallery"]

track_default_position: dict[str, TrackPosition] = {
    "1d-heatmap": "top",
    "2d-rectangle-domains": "center",
    "bar": "top",
    "bedlike": "top",
    "chromosome-labels": "top",
    "gene-annotations": "top",
    "heatmap": "center",
    "horizontal-1d-heatmap": "top",
    "horizontal-bar": "top",
    "horizontal-chromosome-labels": "top",
    "horizontal-gene-annotations": "top",
    "horizontal-heatmap": "top",
    "horizontal-line": "top",
    "horizontal-multivec": "top",
    "left-axis": "left",
    "line": "top",
    "osm-tiles": "center",
    "top-axis": "top",
    "viewport-projection-center": "center",
    "viewport-projection-horizontal": "top",
}

datatype_default_track = {
    "2d-rectangle-domains": "2d-rectangle-domains",
    "bedlike": "bedlike",
    "chromsizes": "horizontal-chromosome-labels",
    "gene-annotations": "horizontal-gene-annotations",
    "matrix": "heatmap",
    "multivec": "horizontal-multivec",
    "vector": "horizontal-bar",
}


def uid() -> str:
    return str(uuid.uuid4())


T = TypeVar("T")


def ensure_list(x: T | list[T] | None) -> list[T]:
    if x is None:
        return []
    return x if isinstance(x, list) else [x]


ModelT = TypeVar("ModelT", bound=BaseModel)


def copy_unique(model: ModelT) -> ModelT:
    """Creates a deep copy of a pydantic BaseModel with new UID."""
    copy = model.__class__(**model.dict())
    if hasattr(copy, "uid"):
        setattr(copy, "uid", uid())
    return copy



from __future__ import annotations

import pytest
from higlass._utils import copy_unique, ensure_list
from pydantic import BaseModel


def test_copy_unique():
    class Person(BaseModel):
        name: str

    person = Person(name="foo")

    other = copy_unique(person)
    assert other is not person
    assert not hasattr(other, "uid")

    class PersonWithId(Person):
        uid: str

    person_with_id = PersonWithId(name="foo", uid="something")
    other = copy_unique(person_with_id)
    assert other is not person
    assert hasattr(other, "uid")
    assert other.uid != person_with_id.uid


@pytest.mark.parametrize("value", [1, [1, 2], None])
def test_ensure_list(value: int | list[int] | None):
    assert isinstance(ensure_list(value), list)



from __future__ import annotations

import threading
import time

import portpicker
import starlette.applications
import uvicorn


class BackgroundServer:
    _app: starlette.applications.Starlette
    _port: int | None
    _server_thread: threading.Thread | None
    _server: uvicorn.Server | None

    def __init__(self, app: starlette.applications.Starlette):
        self._app = app
        self._port = None
        self._server_thread = None
        self._server = None

    @property
    def app(self) -> starlette.applications.Starlette:
        return self._app

    @property
    def port(self) -> int:
        if self._server_thread is None or self._port is None:
            raise RuntimeError("Server not running.")
        return self._port

    def stop(self):
        if self._server_thread is None:
            return self
        assert self._server is not None

        try:
            # queue exit event and wait for thread to terminate
            self._server.should_exit = True
            self._server_thread.join()
        finally:
            self._server = None
            self._server_thread = None

        return self

    def start(
        self,
        port: int | None = None,
        timeout: int = 1,
        daemon: bool = True,
        log_level: str = "warning",
    ):
        if self._server_thread is not None:
            return self

        config = uvicorn.Config(
            app=self.app,
            port=port or portpicker.pick_unused_port(),
            timeout_keep_alive=timeout,
            log_level=log_level,
        )

        self._port = config.port
        self._server = uvicorn.Server(config=config)
        self._server_thread = threading.Thread(target=self._server.run, daemon=daemon)
        self._server_thread.start()

        # wait for the server to start
        while not self._server.started:
            time.sleep(1e-3)

        return self



from __future__ import annotations

import logging
from errno import ENOENT
from typing import Literal

from fuse import FUSE, FuseOSError, LoggingMixIn, Operations
from simple_httpfs import HttpFs

FsName = Literal["http", "https", "ftp"]

logger = logging.getLogger("hg.fuse")


class MultiHttpFs(LoggingMixIn, Operations):
    def __init__(self, schemas: list[FsName], **kwargs):
        logger.info("Starting FUSE at /")
        assert len(schemas) > 0, "must provide at least one schema"
        self.fs = {schema: HttpFs(schema, **kwargs) for schema in schemas}

    def _deref(self, path: str):
        root, *rest = path.lstrip("/").split("/")
        if len(rest) == 1 and rest[0] == "":
            # path == "/http/", "/https/", "/ftp/"
            raise FuseOSError(ENOENT)
        try:
            fs = self.fs[root]
        except KeyError:
            raise FuseOSError(ENOENT)
        return fs, "/" + "/".join(rest)

    def getattr(self, path: str, fh=None):
        logger.debug("getattr %s", path)
        if path == "/":
            first = next(iter(self.fs.values()))
            return first.getattr("/", fh)
        fs, path = self._deref(path)
        return fs.getattr(path, fh)

    def read(self, path: str, size: int, offset: int, fh):
        logger.debug("read %s", (path, size, offset))
        fs, path = self._deref(path)
        return fs.read(path, size, offset, fh)

    def readdir(self, path, fh):
        logger.debug("readdir %s", path)
        if path[-2:] == "..":
            raise NotADirectoryError(path)
        files = list(self.fs) if path == "/" else []
        return [".", "..", *files]

    def destroy(self, path):
        for fs in self.fs.values():
            fs.destroy(path)


def run(mount_point: str, disk_cache_dir: str):
    fs = MultiHttpFs(
        ["http", "https"],
        disk_cache_size=2**25,
        disk_cache_dir=disk_cache_dir,
        lru_capacity=400,
    )
    FUSE(fs, mount_point, foreground=True)



from __future__ import annotations

import logging
import multiprocessing as mp
import pathlib
import time
from urllib.parse import urlparse

logger = logging.getLogger("hg.fuse")

__all__ = [
    "fuse",
    "FuseProcess",
]


class FuseProcess:
    _mnt_name = "schemas"
    _dircache_name = "cache"

    def __init__(self):
        self._fuse_process: mp.Process | None = None
        self._tmp_dir: pathlib.Path | None = None

    def start(self, tmp_dir: str | pathlib.Path):
        try:
            from ._httpfs import run
        except ImportError as e:
            raise ImportError(
                'Install "fusepy" and "simple-httpfs" to enable FUSE.'
            ) from e

        # no need to restart
        tmp_dir = pathlib.Path(tmp_dir).absolute()
        if self._fuse_process and tmp_dir == self._tmp_dir:
            logger.debug("Skipping start. FUSE running in same directory %s", tmp_dir)
            return

        self.stop()

        assert tmp_dir.is_dir(), f"mount dir doesn't exist: {tmp_dir}"

        mount_point = tmp_dir / self._mnt_name
        disk_cache_dir = tmp_dir / self._dircache_name

        if not mount_point.exists():
            mount_point.mkdir()

        if not disk_cache_dir.exists():
            disk_cache_dir.mkdir()

        logger.info("Starting FUSE mount at %s", mount_point)

        args = (str(mount_point) + "/", str(disk_cache_dir) + "/")
        self._fuse_process = mp.Process(target=run, args=args, daemon=True)
        self._fuse_process.start()

        max_iters = 10
        for i in range(max_iters):
            # wait until http is mounted
            if (mount_point / "http").exists():
                break

            if i == max_iters - 1:
                self.stop()
                raise RuntimeError("Failed to setup FUSE")

            time.sleep(0.5)

        self._tmp_dir = tmp_dir

    def stop(self):
        if self._fuse_process is None:
            return

        logger.info("Stopping FUSE running at %s", self._tmp_dir)

        self._fuse_process.terminate()
        self._fuse_process = None
        self._tmp_dir = None

        # TODO: remove cache and mount dirs?
        # make sure stuff is no longer mounted

    def path(self, href: str):
        if self._tmp_dir is None:
            raise RuntimeError("FUSE processes not started")
        url = urlparse(href)
        return str(
            self._tmp_dir / self._mnt_name / f"{url.scheme}/{url.netloc}{url.path}.."
        )


fuse = FuseProcess()



from __future__ import annotations

import itertools
from bisect import bisect_right
from typing import Iterable, Tuple

GenomicPosition = Tuple[str, int]


class Scale:
    """
    A bidirectional mapping between a composite genomic coordinate system and
    a partition of that coordinate system into a sequence of bins.

    The partition is a sequence of bins of a fixed size, with the exception of
    the last bin in each chromosome, which may be smaller than the fixed size.
    Bins do not cross chromosome boundaries. The scale provides a mapping from
    a genomic coordinate to the index of the bin in which it falls, and an
    inverse mapping from a bin index to the genomic coordinate of the bin's
    start.

    Parameters
    ----------
    chromsizes : dict | list[tuple]
        A dictionary of chromosome names and lengths or a list of tuples
        of chromosome names and lengths.
    binsize : int, optional
        The size of each bin in the partition in bp (default: 1).

    Notes
    -----
    The genomic coordinates are 0-based and the bins of the partition are
    half-open intervals, i.e. the start coordinate of a bin is included in the
    bin, but the end is not.
    """

    def __init__(
        self,
        chromsizes: dict[str, int] | Iterable[tuple[str, int]],
        binsize: int = 1,
    ):
        chromsizes = dict(chromsizes)
        names, lengths = zip(*chromsizes.items())
        lengths_binned = [(length + binsize - 1) // binsize for length in lengths]
        chrom_offsets = list(itertools.accumulate(lengths_binned, initial=0))
        self._chrom_names = names
        self._chrom_offsets = chrom_offsets
        self._chrom_lengths_map = chromsizes
        self._chrom_offsets_map = dict(zip(names, chrom_offsets[:-1]))
        self._binsize = binsize

    @property
    def chromsizes(self) -> dict[str, int]:
        """
        A dictionary of the ordered chromosome names and lengths.
        """
        return self._chrom_lengths_map

    @property
    def binsize(self) -> int:
        """
        The size of each bin in the partition in bp.
        """
        return self._binsize

    def __len__(self) -> int:
        return self._chrom_offsets[-1]

    def __repr__(self) -> str:
        return f"Scale(chromsizes={self.chromsizes}, binsize={self.binsize})"

    def __rich_repr__(self) -> Iterable[tuple]:
        yield "chromsizes", self.chromsizes
        yield "binsize", self.binsize

    def __call__(self, gpos: GenomicPosition) -> int:
        """
        Returns the index of the bin in which the given genomic position falls.
        """
        chrom, pos = gpos
        chrom_offset = self._chrom_offsets_map[chrom]
        clen = self._chrom_lengths_map[chrom]
        pos = max(0, min(pos, clen - 1))
        return chrom_offset + pos // self._binsize

    def invert(self, index: int) -> GenomicPosition:
        """
        Returns the genomic position of the start of the bin at the given index.
        """
        n_bins = self._chrom_offsets[-1]
        index = max(0, min(index, n_bins - 1))
        i = bisect_right(self._chrom_offsets, index)
        chrom = self._chrom_names[i - 1]
        rel_offset = index - self._chrom_offsets[i - 1]
        return chrom, rel_offset * self._binsize

    def rebin(self, binsize: int) -> Scale:
        """
        Returns a new scale with the same genomic coordinate system but a
        different bin size.
        """
        return Scale(self._chrom_lengths_map, binsize)



import pytest
from higlass._display import HTMLRenderer, RendererRegistry, renderers


def test_registry():
    registery = RendererRegistry()

    with pytest.raises(ValueError):
        registery.get()

    def renderer(viewconf: dict, **metadata):
        return {
            "text/plain": "some content",
        }

    registery.register("foo", renderer)

    with pytest.raises(ValueError):
        registery.get()

    with pytest.raises(ValueError):
        registery.enable("blah")

    registery.enable("foo")

    render = registery.get()

    mimebundle = render({})

    assert isinstance(mimebundle, dict)
    assert mimebundle["text/plain"] == "some content"


def test_html_renderer():
    """Just a smoke test to make sure our html gets a unique ID"""

    default_renderer = renderers.get()
    assert isinstance(default_renderer, HTMLRenderer)

    mimebundle = default_renderer(
        {
            "foo": "bar",
        }
    )

    assert "text/html" in mimebundle
    assert 'id="jupyter-hg-' in mimebundle["text/html"]
    assert '"foo": "bar"' in mimebundle["text/html"]



from __future__ import annotations

import functools
from typing import Callable, Dict, Optional

from typing_extensions import ParamSpec

from higlass.tilesets import LocalTileset

from ._provider import TilesetProvider, TilesetResource

__all__ = [
    "HgServer",
    "server",
]

P = ParamSpec("P")


class HgServer:
    def __init__(self):
        self._provider: TilesetProvider | None = None
        # We need to keep references to served resources,
        # because the background server uses weakrefs.
        self._tilesets: dict[str, TilesetResource] = {}

    @property
    def port(self):
        if not self._provider:
            raise RuntimeError("Server not started.")
        return self._provider.port

    def reset(self) -> None:
        if self._provider is not None:
            self._provider.stop()
        self._resources = {}

    def enable_proxy(self):
        try:
            import jupyter_server_proxy
        except ImportError as e:
            raise ImportError(
                'Install "jupyter-server-proxy" to enable server proxying.'
            ) from e
        if not self._provider:
            self._provider = TilesetProvider().start()
        self._provider.proxy = True

    def disable_proxy(self):
        if not self._provider:
            raise RuntimeError("Server not started.")
        self._provider.proxy = False

    def register(
        self, tileset_fn: Callable[P, LocalTileset]
    ) -> Callable[P, TilesetResource]:
        """Register a tileset function for this server.

        This is just a convenience method to avoid the repetition of creating
        a tileset and manually adding the tileset to the server.
        """

        @functools.wraps(tileset_fn)
        def wrapper(*args, **kwargs):
            ts = tileset_fn(*args, **kwargs)
            return self.add(ts)

        return wrapper

    def add(
        self,
        tileset: LocalTileset,
        port: int | None = None,
    ) -> TilesetResource:
        """Add a tileset to the server.

        Note: Only tilesets with new uids are added to the server. If the tileset
              uid matches one already on the server, the existing tileset resource
              is returned. Existing tilesets can only be cleared with
              `HgServer.reset()`.
        """
        if self._provider is None:
            self._provider = TilesetProvider().start(port=port)

        if port is not None and port != self._provider.port:
            self._provider.stop().start(port=port)

        if tileset.uid not in self._tilesets:
            self._tilesets[tileset.uid] = self._provider.create(tileset)

        return self._tilesets[tileset.uid]

    def __rich_repr__(self):
        yield "tilesets", self._tilesets
        try:
            port = self.port
        except RuntimeError:
            port = None
        yield "port", port


server = HgServer()



import functools
import hashlib
import pathlib
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence

from ._utils import TrackType
from .api import track

__all__ = [
    "LocalTileset",
    "RemoteTileset",
    "remote",
    "bigwig",
    "multivec",
    "cooler",
    "hitile",
    "bed2ddb",
]

TileId = str
Tile = Dict[str, Any]
TilesetInfo = Dict[str, Any]

DataType = Literal["vector", "multivec", "matrix"]


@dataclass
class LocalTileset:
    tiles: Callable[[Sequence[TileId]], List[Tile]]
    info: Callable[[], TilesetInfo]
    uid: str
    datatype: Optional[DataType] = None
    name: Optional[str] = None


@dataclass
class RemoteTileset:
    uid: str
    server: str
    name: Optional[str] = None

    def track(self, type_: TrackType, **kwargs):
        t = track(
            type_=type_,
            server=self.server,
            tilesetUid=self.uid,
            **kwargs,
        )
        if self.name:
            t.opts(name=self.name, inplace=True)
        return t


def remote(uid: str, server: str = "https://higlass.io/api/v1", **kwargs):
    return RemoteTileset(uid, server, **kwargs)


def hash_absolute_filepath_as_default_uid(fn: Callable[[str, str], LocalTileset]):
    def wrapper(filepath: str, uid: Optional[str] = None):
        if uid is None:
            abspath = pathlib.Path(filepath).absolute()
            uid = hashlib.md5(str(abspath).encode()).hexdigest()
        return fn(filepath, uid)

    return wrapper


@hash_absolute_filepath_as_default_uid
def bigwig(filepath: str, uid: str):
    try:
        from clodius.tiles.bigwig import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "vector" data-server.'
        )

    return LocalTileset(
        datatype="vector",
        tiles=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=uid,
    )


@hash_absolute_filepath_as_default_uid
def multivec(filepath: str, uid: str):
    try:
        from clodius.tiles.multivec import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "multivec" data-server.'
        )

    return LocalTileset(
        datatype="multivec",
        tiles=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=uid,
    )


@hash_absolute_filepath_as_default_uid
def cooler(filepath: str, uid: str):
    try:
        from clodius.tiles.cooler import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "matrix" data-server.'
        )

    return LocalTileset(
        datatype="matrix",
        tiles=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=uid,
    )


@hash_absolute_filepath_as_default_uid
def hitile(filepath: str, uid: str):
    try:
        from clodius.tiles.hitile import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "vector" data-server.'
        )

    return LocalTileset(
        datatype="vector",
        tiles=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=uid,
    )


@hash_absolute_filepath_as_default_uid
def bed2ddb(filepath: str, uid: str):
    try:
        from clodius.tiles.bed2ddb import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "vector" data-server.'
        )

    return LocalTileset(
        datatype="2d-rectangle-domains",
        tiles=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=uid,
    )



from __future__ import annotations

import higlass as hg
import pytest


@pytest.mark.parametrize(
    "args,expected",
    [
        ("horizontal-line", hg.EnumTrack),
        ("multivec", hg.EnumTrack),
        ("1d-heatmap", hg.EnumTrack),
        ("line", hg.EnumTrack),
        ("point", hg.EnumTrack),
        ("bar", hg.EnumTrack),
        ("divergent-bar", hg.EnumTrack),
        ("stacked-interval", hg.EnumTrack),
        ("gene-annotations", hg.EnumTrack),
        ("linear-2d-rectangle-domains", hg.EnumTrack),
        ("chromosome-labels", hg.EnumTrack),
        ("linear-heatmap", hg.EnumTrack),
        ("1d-value-interval", hg.EnumTrack),
        ("2d-annotations", hg.EnumTrack),
        ("2d-chromosome-annotations", hg.EnumTrack),
        ("2d-chromosome-grid", hg.EnumTrack),
        ("2d-chromosome-labels", hg.EnumTrack),
        ("2d-rectangle-domains", hg.EnumTrack),
        ("2d-tiles", hg.EnumTrack),
        ("arrowhead-domains", hg.EnumTrack),
        ("bedlike", hg.EnumTrack),
        ("cross-rule", hg.EnumTrack),
        ("dummy", hg.EnumTrack),
        ("horizontal-1d-annotations", hg.EnumTrack),
        ("horizontal-1d-heatmap", hg.EnumTrack),
        ("horizontal-1d-tiles", hg.EnumTrack),
        ("horizontal-1d-value-interval", hg.EnumTrack),
        ("horizontal-2d-rectangle-domains", hg.EnumTrack),
        ("horizontal-bar", hg.EnumTrack),
        ("horizontal-chromosome-grid", hg.EnumTrack),
        ("horizontal-chromosome-labels", hg.EnumTrack),
        ("horizontal-divergent-bar", hg.EnumTrack),
        ("horizontal-gene-annotations", hg.EnumTrack),
        ("horizontal-heatmap", hg.EnumTrack),
        ("horizontal-line", hg.EnumTrack),
        ("horizontal-multivec", hg.EnumTrack),
        ("horizontal-point", hg.EnumTrack),
        ("horizontal-rule", hg.EnumTrack),
        ("horizontal-vector-heatmap", hg.EnumTrack),
        ("image-tiles", hg.EnumTrack),
        ("left-axis", hg.EnumTrack),
        ("left-stacked-interval", hg.EnumTrack),
        ("mapbox-tiles", hg.EnumTrack),
        ("osm-2d-tile-ids", hg.EnumTrack),
        ("osm-tiles", hg.EnumTrack),
        ("raster-tiles", hg.EnumTrack),
        ("simple-svg", hg.EnumTrack),
        ("square-markers", hg.EnumTrack),
        ("top-axis", hg.EnumTrack),
        ("top-stacked-interval", hg.EnumTrack),
        ("vertical-1d-annotations", hg.EnumTrack),
        ("vertical-1d-heatmap", hg.EnumTrack),
        ("vertical-1d-tiles", hg.EnumTrack),
        ("vertical-1d-value-interval", hg.EnumTrack),
        ("vertical-2d-rectangle-domains", hg.EnumTrack),
        ("vertical-bar", hg.EnumTrack),
        ("vertical-bedlike", hg.EnumTrack),
        ("vertical-chromosome-grid", hg.EnumTrack),
        ("vertical-chromosome-labels", hg.EnumTrack),
        ("vertical-gene-annotations", hg.EnumTrack),
        ("vertical-heatmap", hg.EnumTrack),
        ("vertical-line", hg.EnumTrack),
        ("vertical-multivec", hg.EnumTrack),
        ("vertical-point", hg.EnumTrack),
        ("vertical-rule", hg.EnumTrack),
        ("vertical-vector-heatmap", hg.EnumTrack),
        ("heatmap", hg.HeatmapTrack),
        ("viewport-projection-center", hg.IndependentViewportProjectionTrack),
        ("combined", hg.PluginTrack),
        (("combined", {"contents": []}), hg.CombinedTrack),
        ("blaaaah", hg.PluginTrack),
    ],
)
def test_creates_correct_track(args: str | tuple, expected: hg.Track):
    track_type, kwargs = (args, {}) if isinstance(args, str) else args
    track = hg.track(track_type, **kwargs)  # type: ignore
    assert isinstance(track, expected)  # type: ignore
    assert track.type == track_type


def test_viewport_projection():
    v1 = hg.view(hg.track("heatmap"))
    v2 = hg.view(hg.track("heatmap"))

    v3 = v1.project(v2, on="top")
    assert v3.tracks.top and v3.tracks.top[0].type == "viewport-projection-horizontal"

    v4 = v2.project(v1, on="center", inplace=True)
    assert v2 == v4
    assert v2.tracks.center and v2.tracks.center[1].type == "viewport-projection-center"


def test_hconcat_views():
    v1 = hg.view(hg.track("heatmap"), width=4, height=4)
    v2 = hg.view(hg.track("heatmap"), width=5, height=4)

    viewconf = hg.hconcat(v1, v2)
    assert isinstance(viewconf, hg.Viewconf)
    assert viewconf.views and len(viewconf.views) == 2
    assert viewconf.views[0].layout.x == 0
    assert viewconf.views[0].layout.y == 0
    assert viewconf.views[1].layout.x == 4
    assert viewconf.views[1].layout.y == 0


def test_vconcat_views():
    v1 = hg.view(hg.track("heatmap"), width=4, height=4)
    v2 = hg.view(hg.track("heatmap"), width=5, height=4)

    viewconf = hg.vconcat(v1, v2)
    assert isinstance(viewconf, hg.Viewconf)
    assert viewconf.views and len(viewconf.views) == 2
    assert viewconf.views[0].layout.x == 0
    assert viewconf.views[0].layout.y == 0
    assert viewconf.views[1].layout.x == 0
    assert viewconf.views[1].layout.y == 4


def test_concat_viewconfs():
    v1 = hg.view(hg.track("heatmap"), width=4, height=4).viewconf()
    v2 = hg.view(hg.track("heatmap"), width=5, height=4)

    viewconf = hg.vconcat(v1, v2)
    assert isinstance(viewconf, hg.Viewconf)
    assert viewconf.views and len(viewconf.views) == 2
    assert viewconf.views[0].layout.x == 0
    assert viewconf.views[0].layout.y == 0
    assert viewconf.views[1].layout.x == 0
    assert viewconf.views[1].layout.y == 4

    hg.view(hg.track("heatmap"), width=5, height=3).viewconf()


def test_lock():
    # empty lock throws
    with pytest.raises(AssertionError):
        hg.lock()

    views = [hg.view(hg.track("heatmap")) for _ in range(3)]

    lock = hg.lock(*views, uid="foo")

    assert isinstance(lock, hg.Lock)
    assert lock.uid == "foo"

    for (uid, v), view in zip(lock, views):
        assert uid == view.uid
        assert v == (1, 1, 1)


def test_value_scale_lock():
    # empty lock throws
    with pytest.raises(AssertionError):
        hg.lock()

    tracks = [hg.track("heatmap") for _ in range(3)]
    views = [hg.view(t) for t in tracks]

    lock = hg.lock(*zip(views, tracks), uid="foo")

    assert isinstance(lock, hg.ValueScaleLock)
    assert lock.uid == "foo"

    for (uid, v), (view, track) in zip(lock, zip(views, tracks)):
        assert uid == f"{view.uid}.{track.uid}"
        assert v["track"] == track.uid
        assert v["view"] == view.uid


def test_properties_mixin():
    track = hg.track("heatmap")
    other = track.properties(
        height=500,
    )
    assert track.uid != other.uid
    assert track.height is None
    assert other.height == 500

    other = track.properties(width=400, inplace=True)
    assert track is other
    assert other.uid == track.uid
    assert track.width == 400


def test_options_mixin():
    track = hg.track("heatmap")
    other = track.opts(
        foo="bar",
    )
    assert track.uid != other.uid
    assert track.options is None
    assert other.options and other.options["foo"] == "bar"

    other = track.opts(foo="bar", inplace=True)
    assert track is other
    assert track.uid == other.uid
    assert track.options and track.options["foo"] == "bar"



import pytest
from higlass._scale import Scale


@pytest.fixture
def scale():
    chromsizes = [("chr1", 10000), ("chr2", 20000), ("chr3", 30000)]
    binsize = 1000
    return Scale(chromsizes, binsize)


def test_chromsizes(scale):
    assert scale.chromsizes == {
        "chr1": 10000,
        "chr2": 20000,
        "chr3": 30000,
    }


def test_binsize(scale):
    assert scale.binsize == 1000


def test_len(scale):
    assert len(scale) == 60


def test_repr(scale):
    assert repr(scale) == (
        "Scale(chromsizes={'chr1': 10000, 'chr2': 20000, 'chr3': 30000}, "
        "binsize=1000)"
    )


def get_data():
    # (gpos, offset, binstart)
    return [
        (("chr1", 0), 0, ("chr1", 0)),
        (("chr1", 900), 0, ("chr1", 0)),
        (("chr1", 1000), 1, ("chr1", 1000)),
        (("chr1", 9000), 9, ("chr1", 9000)),
        (("chr1", 9999), 9, ("chr1", 9000)),
        (("chr1", 10000), 9, ("chr1", 9000)),
        (("chr1", 100000), 9, ("chr1", 9000)),
        (("chr2", 0), 10, ("chr2", 0)),
        (("chr2", 900), 10, ("chr2", 0)),
        (("chr2", 1000), 11, ("chr2", 1000)),
        (("chr2", 19000), 29, ("chr2", 19000)),
        (("chr2", 19999), 29, ("chr2", 19000)),
        (("chr2", 20000), 29, ("chr2", 19000)),
        (("chr2", 100000), 29, ("chr2", 19000)),
        (("chr3", 0), 30, ("chr3", 0)),
        (("chr3", 900), 30, ("chr3", 0)),
        (("chr3", 1000), 31, ("chr3", 1000)),
        (("chr3", 29000), 59, ("chr3", 29000)),
        (("chr3", 29999), 59, ("chr3", 29000)),
        (("chr3", 30000), 59, ("chr3", 29000)),
        (("chr3", 100000), 59, ("chr3", 29000)),
    ]


@pytest.mark.parametrize("gpos, offset, binstart", get_data())
def test_call(scale, gpos, offset, binstart):
    assert scale(gpos) == offset


@pytest.mark.parametrize("gpos, offset, binstart", get_data())
def test_invert(scale, gpos, offset, binstart):
    assert scale.invert(offset) == binstart


def test_call_out_of_bounds(scale):
    assert scale(("chr1", -1)) == 0
    assert scale(("chr3", 30000)) == 59


def test_invert_out_of_bounds(scale):
    assert scale.invert(-1) == ("chr1", 0)
    assert scale.invert(60) == ("chr3", 29000)


def test_rebin(scale):
    new_scale = scale.rebin(500)
    assert scale.binsize == 1000
    assert len(scale) == 60
    assert new_scale.binsize == 500
    assert len(new_scale) == 120



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("higlass-python")
except PackageNotFoundError:
    __version__ = "uninstalled"

from higlass_schema import *

import higlass.tilesets
from higlass.api import *
from higlass.fuse import fuse
from higlass.server import server
from higlass.tilesets import remote

bigwig = server.register(higlass.tilesets.bigwig)
multivec = server.register(higlass.tilesets.multivec)
cooler = server.register(higlass.tilesets.cooler)
hitile = server.register(higlass.tilesets.hitile)
bed2ddb = server.register(higlass.tilesets.bed2ddb)



from __future__ import annotations

import json
import uuid
from dataclasses import dataclass, field
from typing import Protocol

import jinja2

HTML_TEMPLATE = jinja2.Template(
    """
<!DOCTYPE html>
<html>
  <head>
    <link rel="stylesheet" href="https://esm.sh/higlass@{{ higlass_version }}/dist/hglib.css">
    <script src="https://unpkg.com/requirejs-toggle"></script>
    {% for plugin_url in plugin_urls %}
    <script src="{{ plugin_url }}"></script>
    {% endfor %}
    <script src="https://unpkg.com/requirejs-toggle"></script>
  </head>
  <body>
    <div id="{{ output_div }}"></div>
  </body>
  <script type="module">
    import hglib from "https://esm.sh/higlass@{{ higlass_version }}?deps=react@{{ react_version }},react-dom@{{ react_version }},pixi.js@{{ pixijs_version }}";
    hglib.viewer(
      document.getElementById('{{ output_div }}'),
      {{ viewconf }},
    );
    </script>
</html>
"""  # noqa
)


def viewconf_to_html(
    viewconf: dict,
    higlass_version: str = "1.11",
    react_version: str = "17",
    pixijs_version: str = "6",
    output_div: str = "vis",
    json_kwds: dict | None = None,
    plugin_urls: list[str] | None = None,
):
    """Embed the viewconf into an HTML template.

    The HTML template includes imports for HiGlass and any specified plugins
    and will display an interactive visualization of the embedded viewconf
    when loaded in the web browser.

    Parameters
    ----------

    viewconf : dict
        The top-level HiGlass viewconf.

    higlass_version : str, optional
        The HiGlass client version.

    react_version : str, optional
        The react and react-dom peer dependency versions.

    pixijs_version : str, optional
        The pixi.js perr dependency version.

    output_div : str, optional
        An id for the div which the visualization renders.

    json_kwds : dict, optional
        Additional arguments to pass to `json.dumps`.

    plugin_urls : list[str], optional
        URLs for plugin tracks or data fetchers to be loaded on the page.
    """
    return HTML_TEMPLATE.render(
        viewconf=json.dumps(viewconf, **(json_kwds or {})),
        higlass_version=higlass_version,
        react_version=react_version,
        pixijs_version=pixijs_version,
        output_div=output_div,
        plugin_urls=plugin_urls or [],
    )


class RendererProtocol(Protocol):
    """A callable that converts a dict viewconf into full mimebundle."""

    def __call__(self, viewconf: dict, **metadata) -> dict:
        """Creates a full mimebundle from a dictionary viewconf.

        Parameters
        ----------

        viewconf : dict
            A top-level HiGlass viewconf (e.g., `hg.Viewconf.dict()`))

        **metadata : dict
            Any extra metadata for the renderer.

        Returns
        -------

        mimebundle : a full mimebundle containing the mapping from all
            mimetypes to data. See return type of `_repr_mimebundle_`
            method in IPython documentation.


        Examples
        --------

        >>> def render_text(viewconf: dict):
        >>>     return {
        >>>         "text/plain": "Viewconf(...)",
        >>>     }

        """
        ...


class BaseRenderer(RendererProtocol):
    def __init__(self, output_div: str = "jupyter-hg-{}", **kwargs):
        self._output_div = output_div
        self.kwargs = kwargs

    @property
    def output_div(self):
        """Creates a unique id for the output div each time it is accessed."""
        return self._output_div.format(uuid.uuid4().hex)

    def __call__(self, viewconf: dict, **metadata):
        raise NotImplementedError()


class HTMLRenderer(BaseRenderer):
    def __call__(self, viewconf: dict, **metadata):
        kwargs = self.kwargs.copy()
        kwargs.update(metadata)
        html = viewconf_to_html(viewconf=viewconf, output_div=self.output_div, **kwargs)
        return {"text/html": html}


@dataclass
class RendererRegistry:
    """A registery for multiple HiGlass renderers.

    Allows for multiple renders to be registered, and dynamically enabled/disabled
    for the viewconf.

    Examples
    --------

    >>> import higlass as hg
    >>> hg.renderers.register("my-renderer", MyRenderer())
    >>> hg.renderers.enable("my-renderer")
    >>> hg.view(hg.track("heatmap")) # calls the custom renderer

    """

    renderers: dict[str, RendererProtocol] = field(default_factory=dict)
    active: str | None = None

    def register(self, name: str, renderer: RendererProtocol):
        """Reigster a custom renderer.

        Parameters
        ----------

        name : str
            A uniquely identifiable name.

        renderer : RendererProtocol
            A Callable that implements the RendererProtocol.
        """
        self.renderers[name] = renderer

    def enable(self, name: str):
        """Examples a previously registered renderer.

        Parameters
        ----------

        name : str
            The name of a previously registered renderer.
        """
        if name not in self.renderers:
            raise ValueError(f"Renderer '{name}' has not been registered.")
        self.active = name

    def get(self):
        """Get the enabled renderer."""
        if self.active is None:
            raise ValueError("No renderer enabled.")
        return self.renderers[self.active]


html_renderer = HTMLRenderer()

renderers = RendererRegistry()
renderers.register("default", html_renderer)
renderers.register("html", html_renderer)
renderers.enable("default")
